from typing import Optional

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin, OneToOneFeatureMixin
from sklearn.utils.validation import check_is_fitted, validate_data


class PointScaler(TransformerMixin, OneToOneFeatureMixin, BaseEstimator):
    """
    A transformer that scales the input data by the intensity value at a given point.
    The point can be specified by an index or by a wavenumber.

    Parameters
    ----------
    point : int,
        The point to scale the data by. It can be an index or a wavenumber.

    wavenumber : array-like, optional
        The wavenumbers of the input data. If not provided, the indices will be used
        instead. Default is None. If provided, the wavenumbers must be provided in
        ascending order.

    Attributes
    ----------
    point_index_ : int
        The index of the point to scale the data by. It is 0 if the wavenumbers are not provided.

    Methods
    -------
    fit(X, y=None)
        Fit the transformer to the input data.

    transform(X, y=0, copy=True)
        Transform the input data by scaling by the value at a given Point.
    """

    def __init__(self, point: int = 0, wavenumbers: Optional[np.ndarray] = None):
        self.point = point
        self.wavenumbers = wavenumbers

    def fit(self, X: np.ndarray, y=None) -> "PointScaler":
        """
        Fit the transformer to the input data.

        Parameters
        ----------
        X : np.ndarray of shape (n_samples, n_features)
            The input data to fit the transformer to.

        y : None
            Ignored.

        Returns
        -------
        self : PointScaler
            The fitted transformer.
        """
        # Check that X is a 2D array and has only finite values
        X = validate_data(
            self, X, y="no_validation", ensure_2d=True, reset=True, dtype=np.float64
        )
        # Set the point index
        if self.wavenumbers is None:
            self.point_index_ = self.point
        else:
            self.point_index_ = self._find_index(self.point)

        return self

    def transform(self, X: np.ndarray, y=None) -> np.ndarray:
        """
        Transform the input data by scaling by the value at a given Point.

        Parameters
        ----------
        X : np.ndarray of shape (n_samples, n_features)
            The input data to transform.

        y : None
            Ignored.

        Returns
        -------
        X_ : np.ndarray of shape (n_samples, n_features)
            The transformed data.
        """
        # Check that the estimator is fitted
        check_is_fitted(self, "point_index_")

        # Check that X is a 2D array and has only finite values
        X_ = validate_data(
            self,
            X,
            y="no_validation",
            ensure_2d=True,
            copy=True,
            reset=False,
            dtype=np.float64,
        )

        # Check that the number of features is the same as the fitted data
        if X_.shape[1] != self.n_features_in_:
            raise ValueError(
                f"Expected {self.n_features_in_} features but got {X_.shape[1]}"
            )

        # Scale the data by Point
        for i, x in enumerate(X_):
            X_[i] = x / x[self.point_index_]

        return X_.reshape(-1, 1) if X_.ndim == 1 else X_

    def _find_index(self, target: float) -> int:
        wavenumbers = np.array(self.wavenumbers)
        return int(np.argmin(np.abs(wavenumbers - target)))



from typing import Literal, Optional

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin, OneToOneFeatureMixin
from sklearn.utils.validation import check_is_fitted, validate_data


class AddNoise(TransformerMixin, OneToOneFeatureMixin, BaseEstimator):
    """
    Add normal noise to the input data.
    """

    def __init__(
        self,
        noise_distribution: Literal["gaussian", "poisson", "exponential"] = "gaussian",
        scale: float = 0.0,
        random_state: Optional[int] = None,
    ):
        self.noise_distribution = noise_distribution
        self.scale = scale
        self.random_state = random_state

    def fit(self, X: np.ndarray, y=None) -> "AddNoise":
        """
        Fit the transformer to the input data.

        Parameters
        ----------
        X : np.ndarray of shape (n_samples, n_features)
            The input data to fit the transformer to.

        y : None
            Ignored.

        Returns
        -------
        self : NormalNoise
            The fitted transformer.
        """
        # Check that X is a 2D array and has only finite values
        X = validate_data(
            self, X, y="no_validation", ensure_2d=True, reset=True, dtype=np.float64
        )

        # Set the number of features
        self.n_features_in_ = X.shape[1]

        # Set the fitted attribute to True
        self._is_fitted = True

        # Instantiate the random number generator
        self._rng = np.random.default_rng(self.random_state)

        return self

    def transform(self, X: np.ndarray, y=None) -> np.ndarray:
        """
        Transform the input data by adding random normal noise.

        Parameters
        ----------
        X : np.ndarray of shape (n_samples, n_features)
            The input data to transform.

        y : None
            Ignored.

        Returns
        -------
        X_ : np.ndarray of shape (n_samples, n_features)
            The transformed data.
        """
        # Check that the estimator is fitted
        check_is_fitted(self, "_is_fitted")

        # Check that X is a 2D array and has only finite values
        X_ = validate_data(
            self,
            X,
            y="no_validation",
            ensure_2d=True,
            copy=True,
            reset=False,
            dtype=np.float64,
        )

        # Check that the number of features is the same as the fitted data
        if X_.shape[1] != self.n_features_in_:
            raise ValueError(
                f"Expected {self.n_features_in_} features but got {X_.shape[1]}"
            )

        # Calculate the standard normal variate
        for i, x in enumerate(X_):
            match self.noise_distribution:
                case "gaussian":
                    X_[i] = self._add_gaussian_noise(x)
                case "poisson":
                    X_[i] = self._add_poisson_noise(x)
                case "exponential":
                    X_[i] = self._add_exponential_noise(x)
                case _:
                    raise ValueError("Invalid noise distribution")

        return X_.reshape(-1, 1) if X_.ndim == 1 else X_

    def _add_gaussian_noise(self, x) -> np.ndarray:
        return x + self._rng.normal(0, self.scale, size=x.shape)

    def _add_poisson_noise(self, x) -> np.ndarray:
        return self._rng.poisson(x, size=x.shape) * self.scale

    def _add_exponential_noise(self, x) -> np.ndarray:
        return x + self._rng.exponential(self.scale, size=x.shape)



from typing import Optional

import numpy as np
from sklearn.base import BaseEstimator
from sklearn.feature_selection._base import SelectorMixin
from sklearn.utils.validation import check_is_fitted, validate_data


class RangeCut(SelectorMixin, BaseEstimator):
    """
    A selector that cuts the input data to a specified range. The range is specified:
    - by the indices of the start and end of the range,
    - by the wavenumbers of the start and end of the range. In this case, the wavenumbers
        must be provided to the transformer when it is initialised. If the wavenumbers
        are not provided, the indices will be used instead. The wavenumbers must be
        provided in ascending order.

    Parameters
    ----------
    start : int, optional
        The index or wavenumber of the start of the range. Default is 0.

    end : int, optional
        The index or wavenumber of the end of the range. Default is -1.

    wavenumbers : array-like, optional
        The wavenumbers of the input data. If not provided, the indices will be used
        instead. Default is None. If provided, the wavenumbers must be provided in
        ascending order.

    Attributes
    ----------
    start_index_ : int
        The index of the start of the range. It is 0 if the wavenumbers are not provided.

    end_index_ : int
        The index of the end of the range. It is -1 if the wavenumbers are not provided.

    wavenuumbers_ : array-like
        The cut wavenumbers of the input data.

    Methods
    -------
    fit(X, y=None)
        Fit the transformer to the input data.
    """

    def __init__(
        self,
        start: int = 0,
        end: int = -1,
        wavenumbers: Optional[np.ndarray] = None,
    ):
        self.start = start
        self.end = end
        self.wavenumbers = wavenumbers

    def fit(self, X: np.ndarray, y=None) -> "RangeCut":
        """
        Fit the transformer to the input data.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input data to fit the transformer to.

        y : None
            Ignored.

        Returns
        -------
        self : RangeCut
            The fitted transformer.
        """
        # Check that X is a 2D array and has only finite values
        X = validate_data(
            self, X, y="no_validation", ensure_2d=True, reset=True, dtype=np.float64
        )
        # Set the start and end indices
        if self.wavenumbers is None:
            self.start_index_ = self.start
            self.end_index_ = self.end
            self.wavenumbers_ = None
        else:
            self.start_index_ = self._find_index(self.start)
            self.end_index_ = self._find_index(self.end)
            self.wavenumbers_ = self.wavenumbers[self.start_index_ : self.end_index_]

        return self

    def _get_support_mask(self):
        """
        Get the boolean mask indicating which features are selected.

        Returns
        -------
        mask : np.ndarray of shape (n_features,)
            The boolean mask indicating which features are selected.
        """
        # Check that the estimator is fitted
        check_is_fitted(self, ["start_index_", "end_index_"])

        # Create the mask
        mask = np.zeros(self.n_features_in_, dtype=bool)
        mask[self.start_index_ : self.end_index_] = True

        return mask

    def _find_index(self, target: float) -> int:
        wavenumbers = np.array(self.wavenumbers)
        return int(np.argmin(np.abs(wavenumbers - target)))



from typing import Optional

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin, OneToOneFeatureMixin
from sklearn.utils.validation import check_is_fitted, validate_data


class ConstantBaselineCorrection(TransformerMixin, OneToOneFeatureMixin, BaseEstimator):
    """
    A transformer that corrects a baseline by subtracting a constant value.
    The constant value is taken by the mean of the features between the start
    and end indices. This is a common preprocessing technique for UV-Vis spectra.

    Parameters
    ----------
    wavenumbers : np.ndarray, optional
        The wavenumbers corresponding to each feature in the input data.

    start : int, optional
        The index of the first feature to use for the baseline correction.

    end : int, optional
        The index of the last feature to use for the baseline correction.

    Attributes
    ----------
    start_index_ : int
        The index of the start of the range. It is 0 if the wavenumbers are not provided.

    end_index_ : int
        The index of the end of the range. It is 1 if the wavenumbers are not provided.

    Methods
    -------
    fit(X, y=None)
        Fit the transformer to the input data.

    transform(X, y=0, copy=True)
        Transform the input data by subtracting the constant baseline value.
    """

    def __init__(
        self,
        start: int = 0,
        end: int = 1,
        wavenumbers: Optional[np.ndarray] = None,
    ) -> None:
        self.start = start
        self.end = end
        self.wavenumbers = wavenumbers

    def fit(self, X: np.ndarray, y=None) -> "ConstantBaselineCorrection":
        """
        Fit the transformer to the input data.

        Parameters
        ----------
        X : np.ndarray of shape (n_samples, n_features)
            The input data to fit the transformer to.

        y : None
            Ignored.

        Returns
        -------
        self : ConstantBaselineCorrection
            The fitted transformer.
        """
        # Check that X is a 2D array and has only finite values
        X = validate_data(
            self, X, y="no_validation", ensure_2d=True, reset=True, dtype=np.float64
        )

        # Set the start and end indices
        if self.wavenumbers is None:
            self.start_index_ = self.start
            self.end_index_ = self.end
        else:
            self.start_index_ = self._find_index(self.start)
            self.end_index_ = self._find_index(self.end)

        return self

    def transform(self, X: np.ndarray, y=0, copy=True) -> np.ndarray:
        """
        Transform the input data by subtracting the constant baseline value.

        Parameters
        ----------
        X : np.ndarray of shape (n_samples, n_features)
            The input data to transform.

        y : int or None, optional
            Ignored.

        copy : bool, optional
            Whether to copy the input data before transforming it.

        Returns
        -------
        X_ : np.ndarray of shape (n_samples, n_features)
            The transformed input data.
        """
        # Check that the estimator is fitted
        check_is_fitted(self, "n_features_in_")

        # Check that X is a 2D array and has only finite values
        X_ = validate_data(
            self,
            X,
            y="no_validation",
            ensure_2d=True,
            copy=True,
            reset=False,
            dtype=np.float64,
        )

        # Base line correct the spectra
        for i, x in enumerate(X_):
            mean_baseline = np.mean(x[self.start_index_ : self.end_index_ + 1])
            X_[i, :] = x - mean_baseline
        return X_.reshape(-1, 1) if X_.ndim == 1 else X_

    def _find_index(self, target: float) -> int:
        wavenumbers = np.array(self.wavenumbers)
        return np.argmin(np.abs(wavenumbers - target)).astype(int)



from typing import Literal, Optional

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin, OneToOneFeatureMixin
from sklearn.utils.validation import check_is_fitted, validate_data


class ExtendedMultiplicativeScatterCorrection(
    TransformerMixin, OneToOneFeatureMixin, BaseEstimator
):
    """Extended multiplicative scatter correction (EMSC) is a preprocessing technique for
    removing non linear scatter effects from spectra. It is based on fitting a polynomial
    regression model to the spectrum using a reference spectrum. The reference spectrum
    can be the mean or median spectrum of a set of spectra or a selected reerence.

    Note that this implementation does not include further extensions of the model using
    orthogonal subspace models.

    Parameters
    ----------
    reference : np.ndarray, optional
        The reference spectrum to use for the correction. If None, the mean
        spectrum will be used. The default is None.
    use_mean : bool, optional
        Whether to use the mean spectrum as the reference. The default is True.
    use_median : bool, optional
        Whether to use the median spectrum as the reference. The default is False.
    order : int, optional
        The order of the polynomial to fit to the spectrum. The default is 2.
    weights : np.ndarray, optional
        The weights to use for the weighted EMSC. If None, the standard EMSC
        will be used. The default is None.


    Attributes
    ----------
    reference_ : np.ndarray
        The reference spectrum used for the correction.

    References
    ----------
    Nils Kristian Afseth, Achim Kohler. Extended multiplicative signal correction
    in vibrational spectroscopy, a tutorial, doi:10.1016/j.chemolab.2012.03.004

    Valeria Tafintseva et al. Correcting replicate variation in spectroscopic data by machine learning and
    model-based pre-processing, doi:10.1016/j.chemolab.2021.104350
    """

    ALLOWED_METHODS = ["mean", "median"]

    # TODO: Check method is valid in instantiation. Right now it is check on fit because it breaks the scikitlearn check_estimator()

    def __init__(
        self,
        method: Literal["mean", "median"] = "mean",
        order: int = 2,
        reference: Optional[np.ndarray] = None,
        weights: Optional[np.ndarray] = None,
    ):
        self.method = method
        self.order = order
        self.reference = reference
        self.weights = weights

    def fit(self, X: np.ndarray, y=None) -> "ExtendedMultiplicativeScatterCorrection":
        """
        Fit the transformer to the input data. If no reference is provided, the
        mean or median spectrum will be calculated from the input data.

        Parameters
        ----------
        X : np.ndarray of shape (n_samples, n_features)
            The input data to fit the transformer to.

        y : None
            Ignored.

        Returns
        -------
        self : MultiplicativeScatterCorrection
            The fitted transformer.
        """
        # Check that X is a 2D array and has only finite values
        X = validate_data(
            self, X, y="no_validation", ensure_2d=True, reset=True, dtype=np.float64
        )

        # Check that the length of the reference is the same as the number of features
        if self.reference is not None:
            if len(self.reference) != self.n_features_in_:
                raise ValueError(
                    f"Expected {self.n_features_in_} features in reference but got {len(self.reference)}"
                )

        if self.weights is not None:
            if len(self.weights) != self.n_features_in_:
                raise ValueError(
                    f"Expected {self.n_features_in_} features in weights but got {len(self.weights)}"
                )

        # Set the reference
        if self.reference is not None:
            self.reference_ = np.array(self.reference)
            self.indices_ = self._calculate_indices(self.reference_)
            self.A_ = self._calculate_A(self.indices_, self.reference_)
            self.weights_ = np.array(self.weights)
            return self

        if self.method == "mean":
            self.reference_ = X.mean(axis=0)
            self.indices_ = self._calculate_indices(X[0])
            self.A_ = self._calculate_A(self.indices_, self.reference_)
            self.weights_ = np.array(self.weights)
            return self

        elif self.method == "median":
            self.reference_ = np.median(X, axis=0)
            self.indices_ = self._calculate_indices(X[0])
            self.A_ = self._calculate_A(self.indices_, self.reference_)
            self.weights_ = np.array(self.weights)
            return self

        else:
            raise ValueError(
                f"Invalid method: {self.method}. Must be one of {self.ALLOWED_METHODS}"
            )

    def transform(self, X: np.ndarray, y=None) -> np.ndarray:
        """
        Transform the input data by applying the multiplicative scatter
        correction.

        Parameters
        ----------
        X : np.ndarray of shape (n_samples, n_features)
            The input data to transform.

        y : None
            Ignored.

        Returns
        -------
        X_ : np.ndarray of shape (n_samples, n_features)
            The transformed data.
        """
        # Check that the estimator is fitted
        check_is_fitted(self, "n_features_in_")

        # Check that X is a 2D array and has only finite values
        X_ = validate_data(
            self,
            X,
            y="no_validation",
            ensure_2d=True,
            copy=True,
            reset=False,
            dtype=np.float64,
        )

        if self.weights is None:
            for i, x in enumerate(X_):
                X_[i] = self._calculate_emsc(x)
            return X_.reshape(-1, 1) if X_.ndim == 1 else X_

        if self.weights is not None:
            for i, x in enumerate(X_):
                X_[i] = self._calculate_weighted_emsc(x)
            return X_.reshape(-1, 1) if X_.ndim == 1 else X_

    def _calculate_weighted_emsc(self, x):
        reg = np.linalg.lstsq(
            np.diag(self.weights_) @ self.A_, x * self.weights_, rcond=None
        )[0]
        x_ = (x - np.dot(self.A_[:, 0:-1], reg[0:-1])) / reg[-1]
        return x_

    def _calculate_emsc(self, x):
        reg = np.linalg.lstsq(self.A_, x, rcond=None)[0]
        x_ = (x - np.dot(self.A_[:, 0:-1], reg[0:-1])) / reg[-1]
        return x_

    def _calculate_indices(self, reference):
        return np.linspace(0, len(reference) - 1, len(reference))

    def _calculate_A(self, indices, reference):
        return np.vstack(
            [[np.power(indices, o) for o in range(self.order + 1)], reference]
        ).T



from ._air_pls import AirPls
from ._ar_pls import ArPls
from ._constant_baseline_correction import ConstantBaselineCorrection
from ._cubic_spline_correction import CubicSplineCorrection
from ._linear_correction import LinearCorrection
from ._non_negative import NonNegative
from ._polynomial_correction import PolynomialCorrection
from ._subtract_reference import SubtractReference


__all__ = [
    "AirPls",
    "ArPls",
    "ConstantBaselineCorrection",
    "CubicSplineCorrection",
    "LinearCorrection",
    "NonNegative",
    "PolynomialCorrection",
    "SubtractReference",
]



import logging
import numpy as np
from scipy.sparse import csc_matrix, eye, diags
from scipy.sparse.linalg import spsolve
from sklearn.base import BaseEstimator, TransformerMixin, OneToOneFeatureMixin
from sklearn.utils.validation import check_is_fitted, validate_data

logger = logging.getLogger(__name__)


class AirPls(TransformerMixin, OneToOneFeatureMixin, BaseEstimator):
    """
    This class implements the AirPLS (Adaptive Iteratively Reweighted Penalized Least Squares) algorithm for baseline
    correction of spectra data. AirPLS is a common approach for removing the baseline from spectra, which can be useful
    in various applications such as spectroscopy and chromatography.

    Parameters
    ----------
    lam : float, optional default=1e2
        The lambda parameter controls the smoothness of the baseline. Increasing the value of lambda results in
        a smoother baseline.

    polynomial_order : int, optional default=1
        The polynomial order determines the degree of the polynomial used to fit the baseline. A value of 1 corresponds
        to a linear fit, while higher values correspond to higher-order polynomials.

    nr_iterations : int, optional default=15
        The number of iterations used to calculate the baseline. Increasing the number of iterations can improve the
        accuracy of the baseline correction, but also increases the computation time.

    Methods
    -------
    fit(X, y=None)
        Fit the estimator to the input data.

    transform(X, y=None)
        Transform the input data by subtracting the baseline.

    _calculate_whittaker_smooth(x, w)
        Calculate the Whittaker smooth of a given input vector x, with weights w.

    _calculate_air_pls(x)
        Calculate the AirPLS baseline of a given input vector x.

    References
    ----------
    - Z.-M. Zhang, S. Chen, and Y.-Z. Liang, Baseline correction using adaptive iteratively reweighted penalized least
      squares. Analyst 135 (5), 1138-1146 (2010).
    """

    def __init__(
        self,
        lam: int = 100,
        polynomial_order: int = 1,
        nr_iterations: int = 15,
    ):
        self.lam = lam
        self.polynomial_order = polynomial_order
        self.nr_iterations = nr_iterations

    def fit(self, X: np.ndarray, y=None) -> "AirPls":
        """Fit the AirPls baseline correction estimator to the input data.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input data.

        y : array-like of shape (n_samples,), optional (default=None)
            The target values.

        Returns
        -------
        self : AirPls
            Returns the instance itself.
        """

        # Check that X is a 2D array and has only finite values
        X = validate_data(
            self, X, y="no_validation", ensure_2d=True, reset=True, dtype=np.float64
        )

        return self

    def transform(self, X: np.ndarray, y=None) -> np.ndarray:
        """Correct the baseline in the input data using the fitted AirPls estimator.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input data.

        y : array-like of shape (n_samples,), optional (default=None)
            The target values.

        Returns
        -------
        X_ : array-like of shape (n_samples, n_features)
            The transformed data with the baseline removed.
        """

        # Check that the estimator is fitted
        check_is_fitted(self, "n_features_in_")

        # Check that X is a 2D array and has only finite values
        X_ = validate_data(
            self,
            X,
            y="no_validation",
            ensure_2d=True,
            copy=True,
            reset=False,
            dtype=np.float64,
        )

        # Calculate the air pls smooth
        for i, x in enumerate(X_):
            X_[i] = x - self._calculate_air_pls(x)

        return X_.reshape(-1, 1) if X_.ndim == 1 else X_

    def _calculate_whittaker_smooth(self, x, w):
        X = np.array(x)
        m = X.size
        E = eye(m, format="csc")
        for i in range(self.polynomial_order):
            E = E[1:] - E[:-1]
        W = diags(w, 0, shape=(m, m))
        A = csc_matrix(W + (self.lam * E.T @ E))
        B = csc_matrix(W @ X.T).toarray().ravel()
        background = spsolve(A, B)
        return np.array(background)

    def _calculate_air_pls(self, x):
        m = x.shape[0]
        w = np.ones(m)

        for i in range(1, self.nr_iterations):
            z = self._calculate_whittaker_smooth(x, w)
            d = x - z
            dssn = np.abs(d[d < 0].sum())

            if dssn < 0.001 * np.abs(x).sum():
                break

            if i == self.nr_iterations - 1:
                break

            w[d >= 0] = 0
            w[d < 0] = np.exp(i * np.abs(d[d < 0]) / dssn)

            negative_d = d[d < 0]
            if negative_d.size > 0:
                w[0] = np.exp(i * negative_d.max() / dssn)

            w[-1] = w[0]

        return z



from typing import Literal, Optional

import numpy as np
from numpy.polynomial import polynomial as poly
from sklearn.base import BaseEstimator, TransformerMixin, OneToOneFeatureMixin
from sklearn.utils.validation import check_is_fitted, validate_data


class IndexShift(TransformerMixin, OneToOneFeatureMixin, BaseEstimator):
    """
    Shift the spectrum a given number of indices between - shift and + shift drawn
    from a discrete uniform distribution.

    Parameters
    ----------
    shift : float, default=0.0
        Shifts the data by a random integer between -shift and shift.

    random_state : int, default=None
        The random state to use for the random number generator.

    Attributes
    ----------
    n_features_in_ : int
        The number of features in the input data.

    _is_fitted : bool
        Whether the transformer has been fitted to data.

    Methods
    -------
    fit(X, y=None)
        Fit the transformer to the input data.

    transform(X, y=0, copy=True)
        Transform the input data by shifting the spectrum.
    """

    def __init__(
        self,
        shift: int = 0,
        fill_method: Literal["constant", "linear", "quadratic"] = "constant",
        random_state: Optional[int] = None,
    ):
        self.shift = shift
        self.fill_method = fill_method
        self.random_state = random_state

    def fit(self, X: np.ndarray, y=None) -> "IndexShift":
        """
        Fit the transformer to the input data.

        Parameters
        ----------
        X : np.ndarray of shape (n_samples, n_features)
            The input data to fit the transformer to.

        y : None
            Ignored.

        Returns
        -------
        self : IndexShift
            The fitted transformer.
        """
        # Check that X is a 2D array and has only finite values
        X = validate_data(
            self, X, y="no_validation", ensure_2d=True, reset=True, dtype=np.float64
        )

        # Set the number of features
        self.n_features_in_ = X.shape[1]

        # Set the fitted attribute to True
        self._is_fitted = True

        # Instantiate the random number generator
        self._rng = np.random.default_rng(self.random_state)

        return self

    def transform(self, X: np.ndarray, y=None) -> np.ndarray:
        """
        Transform the input data by shifting the spectrum.

        Parameters
        ----------
        X : np.ndarray of shape (n_samples, n_features)
            The input data to transform.

        y : None
            Ignored.

        Returns
        -------
        X_ : np.ndarray of shape (n_samples, n_features)
            The transformed data.
        """
        # Check that the estimator is fitted
        check_is_fitted(self, "_is_fitted")

        # Check that X is a 2D array and has only finite values
        X_ = validate_data(
            self,
            X,
            y="no_validation",
            ensure_2d=True,
            copy=True,
            reset=False,
            dtype=np.float64,
        )

        # Check that the number of features is the same as the fitted data
        if X_.shape[1] != self.n_features_in_:
            raise ValueError(
                f"Expected {self.n_features_in_} features but got {X_.shape[1]}"
            )

        # Calculate the standard normal variate
        for i, x in enumerate(X_):
            X_[i] = self._shift_vector(x)

        return X_.reshape(-1, 1) if X_.ndim == 1 else X_

    def _shift_spectrum(self, x) -> np.ndarray:
        shift_amount = self._rng.integers(-self.shift, self.shift, endpoint=True)
        return np.roll(x, shift_amount)

    def _shift_vector(
        self,
        x: np.ndarray,
    ) -> np.ndarray:
        """
        Shift vector with option to fill missing values.

        Args:
            arr: Input numpy array
            shift: Number of positions to shift
            fill_method: Method to fill missing values
                'constant': fill with first/last value
                'linear': fill using linear regression
                'quadratic': fill using quadratic regression

        Returns:
            Shifted numpy array
        """
        shift = self._rng.integers(-self.shift, self.shift, endpoint=True)

        result = np.roll(x, shift)

        if self.fill_method == "constant":
            if shift > 0:
                result[:shift] = x[0]
            elif shift < 0:
                result[shift:] = x[-1]

        elif self.fill_method == "linear":
            if shift > 0:
                x_ = np.arange(5)
                coeffs = poly.polyfit(x_, x[:5], 1)

                extrapolate_x = np.arange(-shift, 0)
                extrapolated_values = poly.polyval(extrapolate_x, coeffs)

                result[:shift] = extrapolated_values

            elif shift < 0:
                x_ = np.arange(5)
                coeffs = poly.polyfit(x_, x[-5:], 1)

                extrapolate_x = np.arange(len(x_), len(x_) - shift)
                extrapolated_values = poly.polyval(extrapolate_x, coeffs)

                result[shift:] = extrapolated_values

        elif self.fill_method == "quadratic":
            if shift > 0:
                # Use first 3 values for quadratic regression
                x_ = np.arange(5)
                coeffs = poly.polyfit(x_, x[:5], 2)

                # Extrapolate to fill shifted region
                extrapolate_x = np.arange(-shift, 0)
                extrapolated_values = poly.polyval(extrapolate_x, coeffs)

                result[:shift] = extrapolated_values

            elif shift < 0:
                # Use last 3 values for quadratic regression
                x_ = np.arange(5)
                coeffs = poly.polyfit(x_, x[-5:], 2)

                # Extrapolate to fill shifted region
                extrapolate_x = np.arange(len(x_), len(x_) - shift)
                extrapolated_values = poly.polyval(extrapolate_x, coeffs)

                result[shift:] = extrapolated_values

        return result



from ._extended_multiplicative_scatter_correction import (
    ExtendedMultiplicativeScatterCorrection,
)
from ._multiplicative_scatter_correction import MultiplicativeScatterCorrection
from ._robust_normal_variate import RobustNormalVariate
from ._standard_normal_variate import StandardNormalVariate

__all__ = [
    "ExtendedMultiplicativeScatterCorrection",
    "MultiplicativeScatterCorrection",
    "RobustNormalVariate",
    "StandardNormalVariate",
]



import os


import pandas as pd
import polars as pl

PACKAGE_DIRECTORY = os.path.dirname(os.path.abspath(__file__))


def load_fermentation_train(set_output="pandas"):
    """
    Loads the training data of the fermentation dataset. This data corresponds to a synthetic dataset measured
    off-line. This dataset is designed to represent the variability of real fermentation data.

    Arguments
    -------
    set_output: str, default='pandas'
        The output format of the data. It can be 'pandas' or 'polars'. If 'polars', the data is returned as a polars DataFrame.

    Returns
    -------
    train_spectra: pd.DataFrame A pandas DataFrame containing the synthetic spectra measured to train the model.
    train_hplc: pd.DataFrame A pandas DataFrame containing the corresponding reference measurements analyzed with HPLC.

    References
    -------
    - Cabaneros Lopez Pau, Udugama Isuru A., Thomsen Sune Tjalfe, Roslander Christian, Junicke Helena,
    Mauricio Iglesias Miguel, Gernaey Krist V. Transforming data into information:
    A parallel hybrid model for real-time state estimation in lignocellulose ethanol fermentations.
    """
    if set_output == "pandas":
        train_spectra = pd.read_csv(PACKAGE_DIRECTORY + "/data/train_spectra.csv")
        train_spectra.columns = train_spectra.columns.astype(float)
        train_hplc = pd.read_csv(PACKAGE_DIRECTORY + "/data/train_hplc.csv")
        return train_spectra, train_hplc

    if set_output == "polars":
        train_spectra = pl.read_csv(PACKAGE_DIRECTORY + "/data/train_spectra.csv")
        train_hplc = pl.read_csv(PACKAGE_DIRECTORY + "/data/train_hplc.csv")
        return train_spectra, train_hplc

    else:
        raise ValueError(
            "Invalid value for set_output. Please use 'pandas' or 'polars'."
        )


def load_fermentation_test(set_output="pandas"):
    """
    Loads the testing data of the fermentation dataset. This data corresponds to real fermentation data measured
    on-line during a fermentation process.

    Arguments
    -------
    set_output: str, default='pandas'
        The output format of the data. It can be 'pandas' or 'polars'. If 'polars', the data is returned as a polars DataFrame.

    Returns
    -------
    test_spectra: pd.DataFrame A pandas DataFrame containing the on-line spectra measured to train the model.
    test_hplc: pd.DataFrame A pandas DataFrame containing the corresponding HPLC measurements.

    References
    -------
    - Cabaneros Lopez Pau, Udugama Isuru A., Thomsen Sune Tjalfe, Roslander Christian, Junicke Helena,
    Mauricio Iglesias Miguel, Gernaey Krist V. Transforming data into information:
    A parallel hybrid model for real-time state estimation in lignocellulose ethanol fermentations.
    """
    if set_output == "pandas":
        fermentation_spectra = pd.read_csv(
            PACKAGE_DIRECTORY + "/data/fermentation_spectra.csv"
        )
        fermentation_spectra.columns = fermentation_spectra.columns.astype(float)
        fermentation_hplc = pd.read_csv(
            PACKAGE_DIRECTORY + "/data/fermentation_hplc.csv"
        )
        return fermentation_spectra, fermentation_hplc

    if set_output == "polars":
        fermentation_spectra = pl.read_csv(
            PACKAGE_DIRECTORY + "/data/fermentation_spectra.csv"
        )
        fermentation_hplc = pl.read_csv(
            PACKAGE_DIRECTORY + "/data/fermentation_hplc.csv"
        )
        return fermentation_spectra, fermentation_hplc

    else:
        raise ValueError(
            "Invalid value for set_output. Please use 'pandas' or 'polars'."
        )


def load_coffee(set_output="pandas"):
    """
    Loads the coffee dataset. This data corresponds to a coffee spectra from three different origins
    measured off-line using attenuated total reflectance Fourier transform infrared spectroscopy (ATR-FTIR).

    Arguments
    -------
    set_output: str, default='pandas'
        The output format of the data. It can be 'pandas' or 'polars'. If 'polars', the data is returned as a polars DataFrame.

    Returns
    -------
    coffee_spectra: pd.DataFrame A pandas DataFrame containing the coffee spectra.
    coffee_labels: pd.DataFrame A pandas DataFrame containing the corresponding labels.
    """
    if set_output == "pandas":
        coffee_spectra = pd.read_csv(PACKAGE_DIRECTORY + "/data/coffee_spectra.csv")
        coffee_labels = pd.read_csv(PACKAGE_DIRECTORY + "/data/coffee_labels.csv")
        return coffee_spectra, coffee_labels

    if set_output == "polars":
        coffee_spectra = pl.read_csv(PACKAGE_DIRECTORY + "/data/coffee_spectra.csv")
        coffee_labels = pl.read_csv(PACKAGE_DIRECTORY + "/data/coffee_labels.csv")
        return coffee_spectra, coffee_labels

    else:
        raise ValueError(
            "Invalid value for set_output. Please use 'pandas' or 'polars'."
        )



import logging
import numpy as np
import scipy.sparse as sp
from scipy.sparse import spdiags, csc_matrix
from scipy.sparse.linalg import splu

from sklearn.base import BaseEstimator, TransformerMixin, OneToOneFeatureMixin
from sklearn.utils.validation import check_is_fitted, validate_data

logger = logging.getLogger(__name__)


class ArPls(TransformerMixin, OneToOneFeatureMixin, BaseEstimator):
    """
    This class implements the Assymmetrically Reweighted Penalized Least Squares (ArPls) is a baseline
    correction method for spectroscopy data. It uses an iterative process
    to estimate and remove the baseline from the spectra.

    Parameters
    ----------
    lam : float, optional (default=1e4)
        The penalty parameter for the difference matrix in the objective function.

    ratio : float, optional (default=0.01)
        The convergence threshold for the weight updating scheme.

    nr_iterations : int, optional (default=100)
        The maximum number of iterations for the weight updating scheme.


    Methods
    -------
    fit(X, y=None)
        Fit the estimator to the data.

    transform(X, y=None)
        Transform the data by removing the baseline.

    _calculate_diff(N)
        Calculate the difference matrix for a given size.

    _calculate_ar_pls(x)
        Calculate the baseline for a given spectrum.

    References
    ----------
    - Sung-June Baek, Aaron Park, Young-Jin Ahn, Jaebum Choo
    Baseline correction using asymmetrically reweighted penalized
    least squares smoothing
    """

    def __init__(
        self,
        lam: float = 1e4,
        ratio: float = 0.01,
        nr_iterations: int = 100,
    ):
        self.lam = lam
        self.ratio = ratio
        self.nr_iterations = nr_iterations

    def fit(self, X: np.ndarray, y=None) -> "ArPls":
        """Fit the estimator to the data.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input data.

        y : array-like of shape (n_samples,), optional (default=None)
            The target values.

        Returns
        -------
        self : ArPls
            Returns the instance itself.
        """

        # Check that X is a 2D array and has only finite values
        X = validate_data(
            self, X, y="no_validation", ensure_2d=True, reset=True, dtype=np.float64
        )

        return self

    def transform(self, X: np.ndarray, y=None) -> np.ndarray:
        """Transform the data by removing the baseline.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input data.

        y : array-like of shape (n_samples,), optional (default=None)
            The target values.

        Returns
        -------
        X_ : array-like of shape (n_samples, n_features)
            The transformed data with the baseline removed.
        """

        # Check that the estimator is fitted
        check_is_fitted(self, "n_features_in_")

        # Check that X is a 2D array and has only finite values
        X_ = validate_data(
            self,
            X,
            y="no_validation",
            ensure_2d=True,
            copy=True,
            reset=False,
        )

        # Calculate the ar pls baseline
        for i, x in enumerate(X_):
            X_[i] = x - self._calculate_ar_pls(x)

        return X_.reshape(-1, 1) if X_.ndim == 1 else X_

    def _calculate_diff(self, N):
        identity_matrix = sp.eye(N, format="csc")
        D2 = sp.diags([1, -2, 1], [0, 1, 2], shape=(N - 2, N), format="csc")
        return D2.dot(identity_matrix).T

    def _calculate_ar_pls(self, x):
        N = len(x)
        D = self._calculate_diff(N)
        H = self.lam * D.dot(D.T)
        w = np.ones(N)
        iteration = 0
        while iteration < self.nr_iterations:
            W = spdiags(w, 0, N, N)
            C = csc_matrix(W + H)
            z = splu(C).solve(w * x)
            d = x - z
            dn = d[d < 0]
            if len(dn) == 0:
                break
            m = np.mean(dn)
            s = np.std(dn)
            exponent = np.clip(2 * (d - (2 * s - m)) / s, -709, 709)
            wt = 1.0 / (1.0 + np.exp(exponent))
            if np.linalg.norm(w - wt) / np.linalg.norm(w) < self.ratio:
                break
            w = wt
            iteration += 1
        return z



import numpy as np
from scipy.sparse import csc_matrix, eye, diags
from scipy.sparse.linalg import spsolve
from sklearn.base import BaseEstimator, TransformerMixin, OneToOneFeatureMixin
from sklearn.utils.validation import (
    check_is_fitted,
    validate_data,
)  # This code is adapted from the following source:

# Z.-M. Zhang, S. Chen, and Y.-Z. Liang,
# Baseline correction using adaptive iteratively reweighted penalized least squares.
# Analyst 135 (5), 1138-1146 (2010).


class WhittakerSmooth(TransformerMixin, OneToOneFeatureMixin, BaseEstimator):
    """
    A transformer that calculates the Whittaker smooth of the input data.

    Parameters
    ----------
    lam : float, optional
        The lambda parameter to use for the Whittaker smooth. Default is 1e2.

    differences : int, optional
        The number of differences to use for the Whittaker smooth. Default is 1.

    Methods
    -------
    fit(X, y=None)
        Fit the transformer to the input data.

    transform(X, y=0, copy=True)
        Transform the input data by calculating the Whittaker smooth.
    """

    def __init__(
        self,
        lam: float = 1e2,
        differences: int = 1,
    ):
        self.lam = lam
        self.differences = differences

    def fit(self, X: np.ndarray, y=None) -> "WhittakerSmooth":
        """
        Fit the transformer to the input data.

        Parameters
        ----------
        X : np.ndarray of shape (n_samples, n_features)
            The input data to fit the transformer to.

        y : None
            Ignored.

        Returns
        -------
        self : WhittakerSmooth
            The fitted transformer.
        """
        # Check that X is a 2D array and has only finite values
        X = validate_data(
            self, X, y="no_validation", ensure_2d=True, reset=True, dtype=np.float64
        )
        return self

    def transform(self, X: np.ndarray, y=None) -> np.ndarray:
        """
        Transform the input data by calculating the Whittaker smooth.

        Parameters
        ----------
        X : np.ndarray of shape (n_samples, n_features)
            The input data to transform.

        y : None
            Ignored.

        Returns
        -------
        X_ : np.ndarray of shape (n_samples, n_features)
            The transformed data.
        """
        # Check that the estimator is fitted
        check_is_fitted(self, "n_features_in_")

        # Check that X is a 2D array and has only finite values
        X_ = validate_data(
            self,
            X,
            y="no_validation",
            ensure_2d=True,
            copy=True,
            reset=False,
            dtype=np.float64,
        )

        # Check that the number of features is the same as the fitted data
        if X_.shape[1] != self.n_features_in_:
            raise ValueError(
                f"Expected {self.n_features_in_} features but got {X_.shape[1]}"
            )

        # Calculate the whittaker smooth
        for i, x in enumerate(X_):
            X_[i] = self._calculate_whittaker_smooth(x)

        return X_.reshape(-1, 1) if X_.ndim == 1 else X_

    def _calculate_whittaker_smooth(self, x):
        X = np.array(x)
        m = X.size
        E = eye(m, format="csc")
        w = np.ones(m)
        for i in range(self.differences):
            E = E[1:] - E[:-1]
        W = diags(w, 0, shape=(m, m))
        A = csc_matrix(W + (self.lam * E.T @ E))
        B = csc_matrix(W @ X.T).toarray().ravel()
        background = spsolve(A, B)
        return np.array(background)



from ._add_noise import AddNoise
from ._baseline_shift import BaselineShift
from ._index_shift import IndexShift
from ._spectrum_scale import SpectrumScale


__all__ = [
    "AddNoise",
    "BaselineShift",
    "IndexShift",
    "SpectrumScale",
]
